import requests

from chaiverse.config import BASE_SUBMITTER_URL, BASE_FEEDBACK_URL
from chaiverse.utils import get_url
from chaiverse.login_cli import get_developer_key_from_cache


class _ChaiverseHTTPClient():
    def __init__(self, developer_key=None, hostname=None, debug_mode=False):
        if developer_key is None:
            developer_key = get_developer_key_from_cache()
        self.developer_key = developer_key
        self.hostname = hostname
        self.debug_mode = debug_mode

    @property
    def headers(self):
        return {"Authorization": f"Bearer {self.developer_key}"}

    def get(self, endpoint, submission_id=None, **kwargs):
        url = get_url(endpoint, hostname=self.hostname, submission_id=submission_id)
        response = self._request(requests.get, url=url, **kwargs)
        return response

    def put(self, endpoint, data):
        url = get_url(endpoint, hostname=self.hostname)
        response = self._request(requests.put, url=url, json=data)
        return response

    def post(self, endpoint, data=None, submission_id=None, **kwargs):
        url = get_url(endpoint, hostname=self.hostname, submission_id=submission_id)
        response = self._request(requests.post, url=url, json=data, **kwargs)
        return response

    def delete(self, endpoint):
        url = get_url(endpoint, hostname=self.hostname)
        response = self._request(requests.delete, url=url)
        return response

    def _request(self, func, url, **kwargs):
        response = func(url=url, headers=self.headers, **kwargs)
        if not self.debug_mode:
            assert response.status_code == 200, response.json()
            response = response.json()
        return response


class SubmitterClient(_ChaiverseHTTPClient):
    def __init__(self,
            developer_key=None,
            hostname=None,
            debug_mode=False):
        hostname = BASE_SUBMITTER_URL if not hostname else hostname
        super().__init__(developer_key, hostname, debug_mode)


class FeedbackClient(_ChaiverseHTTPClient):
    def __init__(self,
            developer_key=None,
            hostname=None,
            debug_mode=False):
        hostname = BASE_FEEDBACK_URL if not hostname else hostname
        super().__init__(developer_key, hostname, debug_mode)



__all__ = ["get_leaderboard"]


import numpy as np
import pandas as pd

from chaiverse.utils import get_submissions, distribute_to_workers
from chaiverse.metrics.feedback_metrics import get_metrics_from_feedback
from chaiverse.metrics.submission_metrics import get_metrics_from_submission
from chaiverse import constants, feedback


def get_leaderboard(
        developer_key=None,
        max_workers=constants.DEFAULT_MAX_WORKERS,
        submission_date_range=None,
        evaluation_date_range=None,
        submission_ids=None,
        fetch_feedback=False,
        ):
    submissions = get_submissions(developer_key, submission_date_range)
    submissions = _filter_submissions_by_submission_ids(submissions, submission_ids) if submission_ids != None else submissions
    submissions = _filter_submissions_by_feedback_count(submissions, constants.PUBLIC_LEADERBOARD_MINIMUM_FEEDBACK_COUNT)
    df = distribute_to_workers(
        get_leaderboard_row,
        submissions.items(),
        developer_key=developer_key,
        evaluation_date_range=evaluation_date_range,
        max_workers=max_workers,
        fetch_feedback=fetch_feedback
    )
    df = pd.DataFrame(df)
    if len(df):
        df.index = np.arange(1, len(df)+1)
    return df


def get_leaderboard_row(submission_item, developer_key=None, evaluation_date_range=None, fetch_feedback=False):
    submission_id, submission_data = submission_item
    submission_data = _fill_submission_with_default_values(submission_id, submission_data)
    if fetch_feedback:
        submission_data = _fill_submission_with_feedback_metrics(submission_id, submission_data, evaluation_date_range, developer_key)
    metrics = get_metrics_from_submission(submission_data) 
    submission = {'submission_id': submission_id, **submission_data, **metrics}
    return submission


def _fill_submission_with_default_values(submission_id, submission_data):
    constant_field_defaults = constants.SUBMISSION_CONSTANT_FIELD_DEFAULTS
    custom_field_defaults = _get_custom_field_defaults(submission_id, submission_data)
    submission = {
        **constant_field_defaults,
        **custom_field_defaults,
        **submission_data
    }
    return submission


def _fill_submission_with_feedback_metrics(submission_id, submission_data, evaluation_date_range, developer_key=None):
    feedback_metrics = {}
    submission_feedback_total = submission_data['double_thumbs_up'] + submission_data['thumbs_up'] + submission_data['thumbs_down']
    is_updated = feedback.is_submission_updated(submission_id, submission_feedback_total)
    feedback_metrics = get_metrics_from_feedback(
        submission_id, 
        developer_key, 
        reload=is_updated, 
        evaluation_date_range=evaluation_date_range
    )
    submission_data = {**submission_data, **feedback_metrics}
    return submission_data


def _get_custom_field_defaults(submission_id, submission):
    custom_field_defaults = {
        'model_name': submission_id
    }
    return custom_field_defaults


def _filter_submissions_by_submission_ids(submissions, submission_ids):
    filtered_submissions = {
        submission_id: data
        for submission_id, data in submissions.items()
        if submission_id in submission_ids
    }
    return filtered_submissions


def _filter_submissions_by_feedback_count(submissions, min_feedback_count):
    submissions = {
        submission_id: submission_data for submission_id, submission_data in submissions.items()
        if submission_data.get('double_thumbs_up', 0) + submission_data['thumbs_up'] + submission_data['thumbs_down'] >= min_feedback_count
    }
    return submissions



__all__ = ["FeedbackMetrics"]


from collections import defaultdict
from typing import Optional

import numpy as np

from chaiverse.lib import date_tools
from chaiverse.feedback import get_feedback
from chaiverse.schemas import DateRange
from chaiverse.metrics.conversation_metrics import ConversationMetrics


class FeedbackMetrics():
    def __init__(self, feedback_data):
        feedback_dict = feedback_data['feedback']
        feedback_dict = _insert_server_epoch_time(feedback_dict)
        self.feedbacks = list(feedback_dict.values())

    def filter_duplicated_uid(self):
        self.feedbacks = _filter_duplicated_uid_feedbacks(self.feedbacks)

    def filter_for_date_range(self, evaluation_date_range: Optional[DateRange]):
        self.feedbacks = [
            feedback for feedback in self.feedbacks
            if date_tools.is_epoch_time_in_date_range(feedback['server_epoch_time'], evaluation_date_range)
        ]

    def calc_metrics(self):
        thumbs_up = self.thumbs_up
        double_thumbs_up = self.double_thumbs_up
        total_feedback_count = self.total_feedback_count
        metrics = {
            'mcl': self.mcl,
            'thumbs_up': thumbs_up,
            'thumbs_down': total_feedback_count - thumbs_up - double_thumbs_up,
            'double_thumbs_up': double_thumbs_up,
            'repetition': self.repetition_score,
        }
        return metrics

    @property
    def convo_metrics(self):
        return [ConversationMetrics(feedback['messages']) for feedback in self.feedbacks]

    @property
    def thumbs_up(self):
        thumbs_ups = [1 if feedback['thumbs_up'] == 1 else 0 for feedback in self.feedbacks]
        thumbs_up = sum(thumbs_ups)
        return thumbs_up

    @property
    def double_thumbs_up(self):
        double_thumbs_ups = [1 if feedback['thumbs_up'] == 2 else 0 for feedback in self.feedbacks]
        double_thumbs_up = sum(double_thumbs_ups)
        return double_thumbs_up

    @property
    def total_feedback_count(self):
        return len(self.feedbacks)

    @property
    def mcl(self):
        return np.mean([m.mcl for m in self.convo_metrics])

    @property
    def repetition_score(self):
        scores = np.array([m.repetition_score for m in self.convo_metrics])
        is_public = np.array([feedback.get('public', True) for feedback in self.feedbacks])
        return np.nanmean(scores[is_public]) if len(scores) else float('nan')


def _insert_server_epoch_time(feedback_dict):
    for feedback_id, feedback in feedback_dict.items():
        feedback['server_epoch_time'] = int(feedback_id.split('_')[-1])
    return feedback_dict


def _filter_duplicated_uid_feedbacks(feedbacks):
    user_feedbacks = defaultdict(list)
    for feedback in feedbacks:
        user_id = feedback["conversation_id"].split("_")[3]
        user_feedbacks[user_id].append(feedback)
    feedbacks = [metrics[0] for _, metrics in user_feedbacks.items()]
    return feedbacks


def get_metrics_from_feedback(submission_id, developer_key, reload=True, evaluation_date_range=None):
    feedback_data = get_feedback(submission_id, developer_key, reload=reload)
    feedback_metrics = FeedbackMetrics(feedback_data.raw_data)
    feedback_metrics.filter_for_date_range(evaluation_date_range)
    feedback_metrics.filter_duplicated_uid()
    metrics = feedback_metrics.calc_metrics()
    return metrics


__all__ = ["display_leaderboard", "display_competition_leaderboard"]

from typing import Optional
import warnings

import pandas as pd
from tabulate import tabulate

from chaiverse.competition_cli import get_competitions
from chaiverse import constants
from chaiverse.metrics.leaderboard_formatter import format_leaderboard
from chaiverse.metrics.leaderboard_api import get_leaderboard
from chaiverse.schemas import Competition
from chaiverse.utils import print_color, cache

pd.set_option('display.max_columns', 50)
pd.set_option('display.max_rows', 500)
pd.set_option('display.width', 500)
pd.set_option("display.colheader_justify","center")

warnings.filterwarnings('ignore', 'Mean of empty slice')


DEFAULT_COMPETITION = Competition(
    display_name='Default',
    leaderboard_format='meval'
)


def display_leaderboard(
    developer_key=None,
    regenerate=False,
    detailed=False,
    max_workers=constants.DEFAULT_MAX_WORKERS,
    leaderboard_format=None
):
    df = display_competition_leaderboard(
        competition=DEFAULT_COMPETITION,
        detailed=detailed,
        regenerate=regenerate,
        developer_key=developer_key,
        max_workers=max_workers,
        leaderboard_format=leaderboard_format
    )
    return df


def display_competition_leaderboard(
    competition:Optional[Competition]=None,
    detailed=False,
    regenerate=False, 
    developer_key=None,
    max_workers=constants.DEFAULT_MAX_WORKERS,
    leaderboard_format=None
):
    competition = competition if competition else get_competitions()[0]
    leaderboard_format = competition.leaderboard_format or 'user' if leaderboard_format is None else leaderboard_format
    fetch_feedback = competition.leaderboard_should_use_feedback

    submission_date_range = competition.submission_date_range
    evaluation_date_range = competition.evaluation_date_range
    submission_ids = competition.enrolled_submission_ids
    display_name = competition.display_name
    title_prefix = 'Detailed ' if detailed else ''
    sort_by = constants.LEADERBOARD_FORMAT_CONFIGS[leaderboard_format]['sort_params']['by']
    display_title = f'{title_prefix}{display_name} Leaderboard (sorted by {sort_by})'

    df = cache(get_leaderboard, regenerate)(
        developer_key=developer_key,
        max_workers=max_workers,
        submission_date_range=submission_date_range,
        evaluation_date_range=evaluation_date_range,
        submission_ids=submission_ids,
        fetch_feedback=fetch_feedback
        )

    if len(df) > 0:
        display_df = df.copy()
        display_df = format_leaderboard(
            display_df, 
            detailed=detailed, 
            format=leaderboard_format
        )
        _pprint_leaderboard(display_df, display_title)
    else:
        print('No eligible submissions found!')
    return df


def _pprint_leaderboard(df, title):
    print_color(f'\n💎 {title}:', 'red')
    print(tabulate(df.round(3).head(30), headers=df.columns, numalign='decimal'))

